// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <inttypes.h>
#include <regex.h>
#include <stdio.h>
#include <perrf.h>
#include "cmd.h"
#include "conf.h"
#include "status.h"
#include "misc.h"

enum TOK_TYPE {
	TOK_NULL = 0,
	TOK_NAME,
	TOK_NUM,
	TOK_REGEX,
	TOK_ASTERISK,
	TOK_HYPHEN,
	TOK_DOLLAR,
	TOK_COLON,
	TOK_ARROBA,
	TOK_UNTERM_REGEX,
	TOK_UNTERM_QUOTE,
};

struct tok {
	unsigned short type;
	size_t siz;
	char *str;
};

static void tok_deinit(struct tok *t)
{
	free(t->str);
}

#define TOK_SIZ_INIT 4096

static void tok_init(struct tok *t)
{
	t->type = TOK_NULL;
	t->siz = TOK_SIZ_INIT;
	t->str = malloc(t->siz);
	if (!t->str)
		abort();
}

inline static void tok_str_realloc(struct tok *t, size_t n)
{
	t->str = realloc(t->str, n);
	if (!t->str)
		abort();
	t->siz = n;
}

inline static void tok_str_ind_set(struct tok *t, size_t i, char c)
{
	if (t->siz <= i)
		tok_str_realloc(t, i + i / 2 + 1);
	t->str[i] = c;
}

static const char *lex(struct tok *t, const char *s)
{
	enum { UNQUTED, DOUBLE_QUOTED, SINGLE_QUOTED } name_state;
	size_t i = 0;
	while (*s && isspace(*s))
		s++;
	if (!*s) {
		t->type = TOK_NULL;
		return NULL;
	}
	while (isdigit(*s))
		tok_str_ind_set(t, i++, *s++);
	if (i) {
		t->type = TOK_NUM;
		goto end;
	}
	if (*s == '/') {
		s++;
		while (*s && *s != '/') {
			if (*s == '\\' && s[1] == '/')
				tok_str_ind_set(t, i++, *(s += 2));
			else
				tok_str_ind_set(t, i++, *s++);
		}
		if (*s) {
			s++;
			t->type = TOK_REGEX;
			goto end;
		}
		t->type = TOK_UNTERM_REGEX;
		goto end;
	}
	if (*s == '*') {
		tok_str_ind_set(t, i++, *s++);
		t->type = TOK_ASTERISK;
		goto end;
	}
	if (*s == '-') {
		tok_str_ind_set(t, i++, *s++);
		t->type = TOK_HYPHEN;
		goto end;
	}
	if (*s == '$') {
		tok_str_ind_set(t, i++, *s++);
		t->type = TOK_DOLLAR;
		goto end;
	}
	if (*s == ':') {
		while (*++s && !isspace(*s))
			tok_str_ind_set(t, i++, *s);
		t->type = TOK_COLON;
		goto end;
	}
	if (*s == '@') {
		while (*++s && !isspace(*s))
			tok_str_ind_set(t, i++, *s);
		t->type = TOK_ARROBA;
		goto end;
	}
	name_state = UNQUTED;
	while (*s && ((!isdigit(*s) && !isspace(*s)
					&& *s != '/' && *s != '*'
					&& *s != '-' && *s != '$'
					&& *s != ':' && *s != '@')
				|| name_state != UNQUTED)) {
		if (*s == '\\')
			tok_str_ind_set(t, i++, *(s+=2));
		else if (name_state == UNQUTED) {
			if (*s == '"') {
				name_state = DOUBLE_QUOTED;
				s++;
			} else if (*s == '\'') {
				name_state = SINGLE_QUOTED;
				s++;
			} else
				tok_str_ind_set(t, i++, *s++);
		} else if (name_state == DOUBLE_QUOTED) {
			if (*s == '"') {
				name_state = UNQUTED;
				s++;
			} else
				tok_str_ind_set(t, i++, *s++);
		} else if (name_state == SINGLE_QUOTED) {
			if (*s == '\'') {
				name_state = UNQUTED;
				s++;
			} else
				tok_str_ind_set(t, i++, *s++);
		}
	}
	t->type = name_state == UNQUTED ? TOK_NAME : TOK_UNTERM_QUOTE;
	goto end;

end:
	tok_str_ind_set(t, i, '\0');
	return s;
}

static void cmd_from(size_t i, struct feed_entry **ep, size_t en)
{
	if (i > en)
		abort();
	char *from = feed_entry_from(ep[i]);
	if (printf("%c %zu %s\n", status(ep[i]), i, from) < 0)
		abort();
	free(from);
	if (fflush(stdout))
		abort();
}

static void cmd_href(size_t i, struct feed_entry **ep, size_t en)
{
	struct feed_link *l = ep[i]->link;
	while (l) {
		if (l->href) {
			if (printf("%s\n", l->href) < 0)
				abort();
		}
		l = l->next;
	}
	if (fflush(stdout))
		abort();

}

static void cmd_id(size_t i, struct feed_entry **ep, size_t en)
{
	if (ep[i]->id) {
		if (printf("%s\n", ep[i]->id) < 0)
			abort();
		if (fflush(stdout))
			abort();
	}
}

static void cmd_view(size_t i, struct feed_entry **ep, size_t en)
{
	struct feed_link *l = ep[i]->link;
	while (l) {
		if (l->href) {
			if (conf_viewer)
				(void)prog_exec(conf_viewer, l->href, NULL);
			else {
				if (printf("%s\n", l->href) < 0)
					abort();
				if (fflush(stdout))
					abort();
			}
		}
		l = l->next;
	}
	status_viewed(ep[i]);
}

static void cmd_feedhref(size_t i, struct feed_entry **ep, size_t en)
{
	struct feed_link *l = ep[i]->feed->link;
	while (l) {
		if (l->href) {
			if (printf("%s\n", l->href) < 0)
				abort();
		}
		l = l->next;
	}
	if (fflush(stdout))
		abort();
}

static void cmd_feeduri(size_t i, struct feed_entry **ep, size_t en)
{
	const char *str = ep[i]->feed->uri;
	if (str) {
		if (printf("%s\n", str) < 0)
			abort();
		if (fflush(stdout))
			abort();
	}
}

static void cmd_summary(size_t i, struct feed_entry **ep, size_t en)
{
	const char *str = ep[i]->summary;
	if (str) {
		if (conf_summary_filter)
			(void)prog_exec(conf_summary_filter, NULL, str);
		else {
			if (printf("%s\n", str) < 0)
				abort();
			if (fflush(stdout))
				abort();
		}
	}
}

const struct cmd cmds[] = {
	{"from", cmd_from},
	{"href", cmd_href},
	{"id", cmd_id},
	{"view", cmd_view},
	{"feeduri", cmd_feeduri},
	{"feedhref", cmd_feedhref},
	{"summary", cmd_summary},
	{"quit", NULL},
	{"exit", NULL},
	{"xit", NULL}
};

const size_t cmdn = sizeof cmds / sizeof cmds[0];

struct cmd_parser {
	struct tok stack[3];
	struct tok la;
	unsigned short n;
	const char *cmd;
	cmd_func func;
	struct feed_entry **ep;
	size_t en;
};

static void cmd_parser_init(struct cmd_parser *cp, const char *cmd,
		struct feed_entry **ep, size_t en)
{
	size_t i;
	for (i = 0; i < sizeof cp->stack / sizeof cp->stack[0]; i++)
		tok_init(&cp->stack[i]);
	tok_init(&cp->la);
	cp->cmd = lex(&cp->la, cmd);
	cp->n = 0;
	cp->func = NULL;
	cp->ep = ep;
	cp->en = en;
}

static void cmd_parser_deinit(struct cmd_parser *cp)
{
	size_t i;
	for (i = 0; i < sizeof cp->stack / sizeof cp->stack[0]; i++)
		tok_deinit(&cp->stack[i]);
	tok_deinit(&cp->la);
}

inline static void tok_swap(struct tok *a, struct tok *b)
{
	struct tok c = *a;
	*a = *b;
	*b = c;
}

inline static int cmd_parser_push(struct cmd_parser *cp)
{
	if (cp->n >= 3)
		abort();
	if (cp->la.type == TOK_NULL)
		return 0;
	tok_swap(&cp->stack[cp->n++], &cp->la);
	cp->cmd = lex(&cp->la, cp->cmd);

	return 1;
}

inline static void cmd_parser_pop(struct cmd_parser *cp)
{
	if (cp->n <= 0)
		abort();
	cp->n--;
}

static int func_reduce(struct cmd_parser *cp)
{
	size_t i;
	if (cp->n != 1 || cp->stack[0].type != TOK_NAME)
		return 0;
	for (i = 0; i < sizeof cmds / sizeof cmds[0]; i++) {
		if (strpref(cp->stack[0].str, cmds[i].name)) {
			cp->func = cmds[i].func;
			cmd_parser_pop(cp);
			return 1;
		}
	}
	perrf("no command named \"%s\"", cp->stack[0].str);

	return 0;
}

inline static int single_reduce(struct cmd_parser *cp)
{
	unsigned long num;
	if (cp->n != 1 || cp->la.type == TOK_HYPHEN)
		return 0;
	if (cp->stack[0].type == TOK_NUM)
		num = strtoul(cp->stack[0].str, NULL, 10);
	else if (cp->stack[0].type == TOK_DOLLAR)
		num = cp->en - 1;
	else
		return 0;
	if (num < cp->en)
		cp->func(num, cp->ep, cp->en);
	cmd_parser_pop(cp);

	return 1;
}

inline static int regex_reduce(struct cmd_parser *cp)
{
	size_t i;
	regex_t re;
	if (cp->n != 1 || cp->stack[0].type != TOK_REGEX)
		return 0;
	if (regcomp(&re, cp->stack[0].str, REG_EXTENDED | REG_NOSUB)) {
		perrf("error parsing regular expression /%s/",
				cp->stack[0].str);
		return 0;
	}
	for (i = 0; i < cp->en; i++) {
		char *from = feed_entry_from(cp->ep[i]);
		if (regexec(&re, from, 0, NULL, 0) == 0)
			cp->func(i, cp->ep, cp->en);
		free(from);
	}
	cmd_parser_pop(cp);

	return 1;
}

inline static int unterm_reduce(struct cmd_parser *cp)
{
	if (cp->n != 1 || (cp->stack[0].type != TOK_UNTERM_REGEX
				&& cp->stack[0].type != TOK_UNTERM_QUOTE))
		return 0;
	perrf("unterminated %s",
			cp->stack[0].type == TOK_UNTERM_REGEX ?
			"regular expression" : "quote");
	cmd_parser_pop(cp);

	return 1;
}

inline static int asterisk_reduce(struct cmd_parser *cp)
{
	size_t i;
	if (cp->n != 1 || cp->stack[0].type != TOK_ASTERISK)
		return 0;
	for (i = 0; i < cp->en; i++)
		cp->func(i, cp->ep, cp->en);
	cmd_parser_pop(cp);

	return 1;
}

inline static int range_reduce(struct cmd_parser *cp)
{
	size_t i, a, b;
	if (cp->n != 3 || cp->stack[0].type != TOK_NUM
			|| cp->stack[1].type != TOK_HYPHEN)
		return 0;
	a = strtoul(cp->stack[0].str, NULL, 10);
	if (cp->stack[2].type == TOK_NUM)
		b = strtoul(cp->stack[2].str, NULL, 10);
	else if (cp->stack[2].type == TOK_DOLLAR)
		b = cp->en - 1;
	else
		return 0;
	for (i = a; i <= b && i < cp->en; i++)
		cp->func(i, cp->ep, cp->en);
	cmd_parser_pop(cp);
	cmd_parser_pop(cp);
	cmd_parser_pop(cp);

	return 1;
}

inline static int title_reduce(struct cmd_parser *cp)
{
	size_t i;
	char *title;
	if (cp->n != 1 || cp->stack[0].type != TOK_NAME)
		return 0;
	for (i = 0; i < cp->en; i++) {
		title = cp->ep[i]->feed->title;
		if (title) {
			if (strcmp(cp->stack[0].str, title) == 0)
				cp->func(i, cp->ep, cp->en);
		}
	}
	cmd_parser_pop(cp);

	return 1;
}

inline static int colon_reduce(struct cmd_parser *cp)
{
	size_t i;
	char st;
	if (cp->n != 1 || cp->stack[0].type != TOK_COLON)
		return 0;
	char str[strlen(cp->stack[0].str) + 1];
	for (i = 0; i < sizeof str / sizeof str[0] - 1; i++)
		str[i] = tolower(cp->stack[0].str[i]);
	str[i] = '\0';
	if (strpref(str, "new"))
		st = STATUS_NEW;
	else if (strpref(str, "unseen"))
		st = STATUS_UNSEEN;
	else if (strpref(str, "viewed"))
		st = STATUS_VIEWED;
	else  {
		perrf("unrecognized status :%s", cp->stack[0].str);
		return 1;
	}
	for (i = 0; i < cp->en; i++) {
		if (st == status(cp->ep[i]))
			cp->func(i, cp->ep, cp->en);
	}
	cmd_parser_pop(cp);

	return 1;
}

inline static int arroba_reduce(struct cmd_parser *cp)
{
	size_t i;
	struct feed_link *l;
	if (cp->n != 1 || cp->stack[0].type != TOK_ARROBA)
		return 0;
	const char * const str = cp->stack[0].str;
	for (i = 0; i < cp->en; i++) {
		if (cp->ep[i]->feed->uri &&
				strcmp(cp->ep[i]->feed->uri, str) == 0)
			cp->func(i, cp->ep, cp->en);
		else {
			l = cp->ep[i]->link;
			while (l) {
				if (l->href && strcmp(l->href, str) == 0) {
					cp->func(i, cp->ep, cp->en);
					break;
				}
				l = l->next;
			}
		}
	}
	cmd_parser_pop(cp);

	return 1;
}

/* Parse and execute command given in string str. */
int cmd(const char *str, struct feed_entry **ep, size_t en)
{
	int res = 0;
	struct cmd_parser cp;
	cmd_parser_init(&cp, str, ep, en);
	if (!cmd_parser_push(&cp))
		goto end;
	if (unterm_reduce(&cp)) {
		res = CMD_RES_ERR;
		goto end;
	}
	if (!func_reduce(&cp)) {
		res = CMD_RES_ERR;
		goto end;
	}
	if (!cp.func) {
		res = CMD_RES_QUIT;
		goto end;
	}
	while (cmd_parser_push(&cp) || cp.n == 3) {
		if (single_reduce(&cp))
			continue;
		if (regex_reduce(&cp))
			continue;
		if (asterisk_reduce(&cp))
			continue;
		if (range_reduce(&cp))
			continue;
		if (title_reduce(&cp))
			continue;
		if (colon_reduce(&cp))
			continue;
		if (arroba_reduce(&cp))
			continue;
		if (unterm_reduce(&cp)) {
			res = CMD_RES_ERR;
			break;
		}
		if (cp.n == 3)
			break;
	}
	if (cp.n) {
		perrf("invalid expression");
		res = CMD_RES_ERR;
		goto end;
	}

end:
	cmd_parser_deinit(&cp);
	return res;
}
