// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <expat.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <perrf.h>
#include <unistd.h>
#include <regex.h>
#include <sprintf2.h>
#include <ctype.h>
#include <hashtab.h>
#include "feed.h"

/* Formats supported:
 * (1) Atom (RFC 4287);
 * (2) RSS 0.91, 0.92 & 2.0 (rssboard.org);
 * (3) RDF Site Summary 1.0 (RSS-DEV). */

#define BUF_SIZ (16*1024*1024)

/* Caller must free returned string. */
static char *uri_escape(const char *uri)
{
	size_t i, n;
	char *res;
	char hex[3];
	for (i = 0, n = 0; uri[i]; i++) {
		if (!isgraph(uri[i]))
			n++;
	}
	if (n == 0) {
		res = strdup(uri);
		if (!res)
			abort();
		return res; 
	}
	res = malloc(i + 2 * n + 1);
	if (!res)
		abort();
	for (i = 0, n = 0; uri[i]; i++) {
		if (!isgraph(uri[i])) {
			const unsigned u = (unsigned char)uri[i];
			if (snprintf(hex, sizeof hex, "%02x", u) != 2)
				abort();
			res[i + n] = '%';
			res[i + ++n] = hex[0];
			res[i + ++n] = hex[1];
		} else
			res[i + n] = uri[i];
	}
	res[i + n] = '\0';

	return res;
}

static struct feed_link *feed_link_new()
{
	struct feed_link *res = malloc(sizeof *res);
	if (!res)
		abort();
	res->rel = FEED_LINK_REL_ALTERNATE;
	res->href = NULL;
	res->lang = NULL;
	res->type = NULL;
	res->next = NULL;

	return res;
}

static void feed_link_free(struct feed_link *link)
{
	while (link) {
		free(link->href);
		free(link->lang);
		void * const p = link;
		link = link->next;
		free(p);
	}
}

static struct feed_link *feed_link_new_attrs(const char **attrs)
{
	struct feed_link *res = feed_link_new();
	while (*attrs) {
		if (strcmp(attrs[0], "href") == 0)
			res->href = uri_escape(attrs[1]);
		else if (strcmp(attrs[0], "hreflang") == 0) {
			if (!(res->lang = strdup(attrs[1])))
				abort();
		} else if (strcmp(attrs[0], "type") == 0) {
			if (!(res->type = strdup(attrs[1])))
				abort();
		} else if (strcmp(attrs[0], "rel") == 0) {
			if (strcmp(attrs[1], "alternate") == 0)
				res->rel = FEED_LINK_REL_ALTERNATE;
			else if (strcmp(attrs[1], "related") == 0)
				res->rel = FEED_LINK_REL_RELATED;
			else if (strcmp(attrs[1], "self") == 0)
				res->rel = FEED_LINK_REL_SELF;
			else if (strcmp(attrs[1], "enclosure") == 0)
				res->rel = FEED_LINK_REL_ENCLOSURE;
			else if (strcmp(attrs[1], "via") == 0)
				res->rel = FEED_LINK_REL_VIA;
			else
				res->rel = FEED_LINK_REL_OTHER;
		}
		attrs += 2;
	}

	return res;
}

static struct feed_entry *feed_entry_new()
{
	struct feed_entry *res = malloc(sizeof *res);
	if (!res)
		abort();
	res->title = NULL;
	res->id = NULL;
	res->summary = NULL;
	res->t = (time_t)-1;
	res->tm = (struct tm){ 0 };
	res->feed = NULL;
	res->link = NULL;
	res->next = NULL;

	return res;
}

static void feed_entry_free(struct feed_entry *entry)
{
	while (entry) {
		free(entry->title);
		free(entry->id);
		free(entry->summary);
		feed_link_free(entry->link);
		void * const p = entry;
		entry = entry->next;
		free(p);
	}
}

/* Returned value must be free'd by the caller. */
char *feed_entry_from(struct feed_entry *entry)
{
	char *str = sprintf2("%02d:%02d %02d-%02d-%04d  %s \t%s",
			entry->tm.tm_hour,
			entry->tm.tm_min,
			entry->tm.tm_mday,
			/* Note that January is 0. */
			entry->tm.tm_mon + 1,
			entry->tm.tm_year + 1900,
			entry->feed->title,
			entry->title ? entry->title :
			entry->summary ? entry->summary :
			"No title");
	if (!str)
		abort();
	return str;
}

static struct feed *feed_new()
{
	struct feed *res = malloc(sizeof *res);
	if (!res)
		abort();
	res->title = NULL;
	res->id = NULL;
	res->uri = NULL;
	res->len = 0;
	res->entry = NULL;
	res->link = NULL;

	return res;
}

void feed_free(struct feed *f)
{
	if (f) {
		free(f->title);
		free(f->id);
		feed_entry_free(f->entry);
		feed_link_free(f->link);
	}
}

void feed_uri_set(struct feed *f, const char *uri)
{
	free(f->uri);
	f->uri = uri_escape(uri);
}

static const char *ns_test(const char *name, const char *namespace)
{
	size_t i;
	for (i = 0; name[i] && namespace[i]; i++) {
		if (namespace[i] != name[i])
			return NULL;
	}
	if (name[i++] != '\xff')
		return NULL;

	return name + i;
}

enum ATOM_TAG {
	ATOM_TAG_FEED = 0,
	ATOM_TAG_TITLE,
	ATOM_TAG_LINK,
	ATOM_TAG_ID,
	ATOM_TAG_ENTRY,
	ATOM_TAG_UPDATED,
	ATOM_TAG_PUBLISHED,
	ATOM_TAG_SUMMARY,
	ATOM_TAG_OTHER
};

static const char atom_ns[] = "http://www.w3.org/2005/Atom";

static int atom_tag_lex(const char *name)
{
	name = ns_test(name, atom_ns);
	if (!name)
		return -1;
	if (strcmp(name, "feed") == 0)
		return ATOM_TAG_FEED;
	else if (strcmp(name, "title") == 0)
		return ATOM_TAG_TITLE;
	else if (strcmp(name, "link") == 0)
		return ATOM_TAG_LINK;
	else if (strcmp(name, "entry") == 0)
		return ATOM_TAG_ENTRY;
	else if (strcmp(name, "id") == 0)
		return ATOM_TAG_ID;
	else if (strcmp(name, "updated") == 0)
		return ATOM_TAG_UPDATED;
	else if (strcmp(name, "published") == 0)
		return ATOM_TAG_PUBLISHED;
	else if (strcmp(name, "summary") == 0)
		return ATOM_TAG_SUMMARY;
	else
		return ATOM_TAG_OTHER;
}

enum RSS_TAG {
	RSS_TAG_RSS = 0,
	RSS_TAG_TITLE,
	RSS_TAG_CHANNEL,
	RSS_TAG_LINK,
	RSS_TAG_PUBDATE,
	RSS_TAG_GUID,
	RSS_TAG_ITEM,
	RSS_TAG_DESCRIPTION,
	RSS_TAG_OTHER
};

static int rss_tag_lex(const char *name)
{
	if (strcmp(name, "rss") == 0)
		return RSS_TAG_RSS;
	else if (strcmp(name, "title") == 0)
		return RSS_TAG_TITLE;
	else if (strcmp(name, "channel") == 0)
		return RSS_TAG_CHANNEL;
	else if (strcmp(name, "link") == 0)
		return RSS_TAG_LINK;
	else if (strcmp(name, "pubDate") == 0)
		return RSS_TAG_PUBDATE;
	else if (strcmp(name, "guid") == 0)
		return RSS_TAG_GUID;
	else if (strcmp(name, "item") == 0)
		return RSS_TAG_ITEM;
	else if (strcmp(name, "description") == 0)
		return RSS_TAG_DESCRIPTION;
	else
		return RSS_TAG_OTHER;
}

enum RDF_TAG {
	RDF_TAG_RDF = 0,
	RDF_TAG_CHANNEL,
	RDF_TAG_ITEM,
	RDF_TAG_TITLE,
	RDF_TAG_LINK,
	RDF_TAG_DESCRIPTION,
	RDF_TAG_DATE,
	RDF_TAG_OTHER
};

const char rdf_ns[] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
const char rdfss_1_0_ns[] = "http://purl.org/rss/1.0/";
const char rdfss_1_0_dc_ns[] = "http://purl.org/dc/elements/1.1/";

static int rdf_tag_lex(const char *name)
{
	size_t i;
	const char *tag = NULL;
	const char *namespaces[] = { rdf_ns, rdfss_1_0_ns, rdfss_1_0_dc_ns };
	for (i = 0; i < sizeof namespaces / sizeof namespaces[0]
			&& !tag; i++)
		tag = ns_test(name, namespaces[i]);
	if (!tag)
		return -1;
	if (strcmp(tag, "RDF") == 0)
		return RDF_TAG_RDF;
	else if (strcmp(tag, "channel") == 0)
		return RDF_TAG_CHANNEL;
	else if (strcmp(tag, "item") == 0)
		return RDF_TAG_ITEM;
	else if (strcmp(tag, "title") == 0)
		return RDF_TAG_TITLE;
	else if (strcmp(tag, "link") == 0)
		return RDF_TAG_LINK;
	else if (strcmp(tag, "description") == 0)
		return RDF_TAG_DESCRIPTION;
	else if (strcmp(tag, "date") == 0)
		return RDF_TAG_DATE;
	else
		return RDF_TAG_OTHER;
}

static int utc_hour_diff;
static int utc_min_diff;

static int utc_diff_init()
{
	struct tm gmtm;
	struct tm ltm;
	tzset();
	const time_t t = time(NULL);
	if (t == (time_t)-1)
		return 1;
	if (!gmtime_r(&t, &gmtm))
		return 1;
	if (!localtime_r(&t, &ltm))
		return 1;
	utc_hour_diff = ltm.tm_hour - gmtm.tm_hour;
	utc_min_diff = ltm.tm_min - gmtm.tm_min;

	return 0;
}

static char xsd_time_date_reg[] =
	"^(-?[0-9]{4,})"
	"-([0-9]{2})"
	"-([0-9]{2})"
	"T([0-9]{2})"
	":([0-9]{2})"
	":([0-9]{2})"
	"(\\.[0-9]+)?"
	"("
		"([-+][0-9]{2})"
		":([0-9]{2})"
		"|"
		"Z"
	")$";
static regex_t xsd_time_date_preg;

static int xsd_time_date_parse(const char *str, struct tm *tm)
{
	regmatch_t pm[11];
	if (regexec(&xsd_time_date_preg, str,
				sizeof pm / sizeof pm[0], pm, 0))
		return 1;
	tm->tm_year = strtol(str + pm[1].rm_so, NULL, 10) - 1900;
	/* Note that January is 0 here. */
	tm->tm_mon = strtol(str + pm[2].rm_so, NULL, 10) - 1;
	tm->tm_mday = strtol(str + pm[3].rm_so, NULL, 10);
	tm->tm_hour = strtol(str + pm[4].rm_so, NULL, 10);
	tm->tm_min = strtol(str + pm[5].rm_so, NULL, 10);
	tm->tm_sec = strtol(str + pm[6].rm_so, NULL, 10);
	if (pm[8].rm_so != -1 && str[pm[8].rm_so] != 'Z') {
		tm->tm_hour -= strtol(str + pm[9].rm_so, NULL, 10);
		tm->tm_min -= strtol(str + pm[10].rm_so, NULL, 10);
	}
	tm->tm_hour += utc_hour_diff;
	tm->tm_min += utc_min_diff;
	tm->tm_isdst = 0;

	return 0;
}

static char rfc822_time_date_reg[] =
	"^"
	"((Mon|Tue|Wed|Thu|Fri|Sat|Sun)[[:space:]]*,)?"
	"[[:space:]]+"
	"([0-9]{1,2})"
	"[[:space:]]+"
	"(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"
	"[[:space:]]+"
	"([0-9]+)"
	"[[:space:]]+"
	"([0-9]{2})"
	"[[:space:]]*:[[:space:]]*"
	"([0-9]{2})"
	"[[:space:]]*"
	"(:[[:space:]]*[0-9]{2})?"
	"[[:space:]]*"
	"(UT|GMT|[ECMP][SD]T|Z|[-+][0-9]{4})"
	"$";
static regex_t rfc822_time_date_preg;

static int rfc822_month(const char *str)
{
	switch (toupper(str[0])) {
		case 'J':
			if (toupper(str[1]) == 'A')
				return 0;
			if (toupper(str[2]) == 'N')
				return 5;
			if (toupper(str[2]) == 'L')
				return 6;
		case 'F':
			return 1;
		case 'M':
			if (toupper(str[2]) == 'R')
				return 2;
			if (toupper(str[2]) == 'Y')
				return 4;
		case 'A':
			if (toupper(str[1]) == 'P')
				return 3;
			if (toupper(str[1]) == 'U')
				return 7;
		case 'S':
			return 8;
		case 'O':
			return 9;
		case 'N':
			return 10;
		case 'D':
			return 11;
	}

	return -1;
}

static int rfc822_timezone(const char *str)
{
	switch (toupper(str[0])) {
		case '-':
		case '+':
			return strtol(str, NULL, 10);
		case 'G':
		case 'U':
		case 'Z':
			return 0;
		case 'E':
			if (toupper(str[1]) == 'S')
				return -500;
			if (toupper(str[1]) == 'D')
				return -400;
			break;
		case 'C':
			if (toupper(str[1] == 'S'))
				return -600;
			if (toupper(str[1] == 'D'))
				return -500;
			break;
		case 'M':
			if (toupper(str[1] == 'S'))
				return -700;
			if (toupper(str[1] == 'D'))
				return -600;
		case 'P':
			if (toupper(str[1] == 'S'))
				return -800;
			if (toupper(str[1] == 'D'))
				return -700;
	}
	return 0;
}

static int rfc822_time_date_parse(const char *str, struct tm *tm)
{
	regmatch_t pm[10];
	if (regexec(&rfc822_time_date_preg, str,
				sizeof pm / sizeof pm[0], pm, 0))
		return 1;
	tm->tm_mday = strtol(str + pm[3].rm_so, NULL, 10);
	tm->tm_mon = rfc822_month(str + pm[4].rm_so);
	if (tm->tm_mon < 0)
		abort();
	tm->tm_year = strtol(str + pm[5].rm_so, NULL, 10);
	if (tm->tm_year >= 100)
		tm->tm_year -= 1900;
	tm->tm_hour = strtol(str + pm[6].rm_so, NULL, 10);
	tm->tm_min = strtol(str + pm[7].rm_so, NULL, 10);
	if (pm[8].rm_so != -1)
		tm->tm_sec = strtol(str + pm[6].rm_so, NULL, 10);
	else
		tm->tm_sec = 0;
	const int tz = rfc822_timezone(str + pm[9].rm_so);
	tm->tm_hour += utc_hour_diff - tz / 100;
	tm->tm_min += utc_min_diff - tz % 100;
	tm->tm_isdst = 0;

	return 0;
}

static char w3cdtf_time_date_reg[] =
	"^(-?[0-9]{4,})"
	"("
		"-([0-9]{2})"
		"("
			"-([0-9]{2})"
			"("
				"T([0-9]{2})"
				":([0-9]{2})"
				"("
					":([0-9]{2})"
					"(\\.[0-9]+)?"
				")?"
				"("
					"([-+][0-9]{2})"
					":([0-9]{2})"
					"|"
					"Z"
				")"
			")?"
		")?"
	")?"
	"$";
static regex_t w3cdtf_time_date_preg;

static int w3cdtf_time_date_parse(const char *str, struct tm *tm)
{
	regmatch_t pm[15];
	*tm = (struct tm){ 0 };
	if (regexec(&w3cdtf_time_date_preg, str,
				sizeof pm / sizeof pm[0], pm, 0))
		return 1;
	tm->tm_year = strtol(str + pm[1].rm_so, NULL, 10) - 1900;
	if (pm[3].rm_so == -1)
		return 0;
	/* Note that January is 0 here. */
	tm->tm_mon = strtol(str + pm[3].rm_so, NULL, 10) - 1;
	if (pm[5].rm_so == -1)
		return 0;
	tm->tm_mday = strtol(str + pm[5].rm_so, NULL, 10);
	if (pm[7].rm_so == -1)
		return 0;
	tm->tm_hour = strtol(str + pm[7].rm_so, NULL, 10);
	tm->tm_min = strtol(str + pm[8].rm_so, NULL, 10);
	if (pm[13].rm_so != -1 && str[pm[12].rm_so] != 'Z') {
		tm->tm_hour -= strtol(str + pm[13].rm_so, NULL, 10);
		tm->tm_min -= strtol(str + pm[14].rm_so, NULL, 10);
	}
	tm->tm_hour += utc_hour_diff;
	tm->tm_min += utc_min_diff;
	if (pm[10].rm_so == -1)
		return 0;
	tm->tm_sec = strtol(str + pm[10].rm_so, NULL, 10);

	return 0;
}

struct feed_parser {
	XML_Parser p;

	int err;

	struct feed *feed;

	size_t skip_depth;
	void (*restore)(struct feed_parser *);

	char **leaf_ret;
	char *leaf_buf;
	size_t leaf_len;

	char *feed_entry_time_str;
};

static void stop(struct feed_parser *fp)
{
	fp->err = 1;
	if (!XML_StopParser(fp->p, XML_FALSE))
		abort();
}

static void leaf_start(void *data, const char *name, const char **attrs)
{
	perrf("element containing children should have been leaf");
	stop(data);
	return;
}

static void leaf_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	/* Note that leaf_char below ensures that fp->leaf_buf
	 * is always big enough to hold fp->leaf_len + 1 chars. */
	fp->leaf_buf[fp->leaf_len] = '\0';
	*fp->leaf_ret = fp->leaf_buf;
	fp->leaf_buf = NULL;
	fp->leaf_len = 0;
	fp->restore(fp);
}

static void leaf_char(void *data, const char *buf, int len)
{
	struct feed_parser *fp = data;
	/* Note we add one here in order to
	 * always be able to add the terminating null
	 * in leaf_end. */
	fp->leaf_buf = realloc(fp->leaf_buf, fp->leaf_len + len + 1);
	if (!fp->leaf_buf)
		abort();
	memcpy(fp->leaf_buf + fp->leaf_len, buf, len);
	fp->leaf_len += len;
}

static void leaf(struct feed_parser *fp, char **leaf_ret,
		void (*restore)(struct feed_parser *))
{
	fp->leaf_ret = leaf_ret;
	fp->restore = restore;
	XML_SetElementHandler(fp->p, leaf_start, leaf_end);
	XML_SetCharacterDataHandler(fp->p, leaf_char);
}

static void skip_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	fp->skip_depth++;
}

static void skip_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (--fp->skip_depth == 0)
		fp->restore(fp);
}

static void skip(struct feed_parser *fp, void (*restore)(struct feed_parser *))
{
	fp->skip_depth = 1;
	fp->restore = restore;
	XML_SetElementHandler(fp->p, skip_start, skip_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void entry_restore(struct feed_parser *fp);

static void entry_time_restore(struct feed_parser *fp)
{
	if (xsd_time_date_parse(fp->feed_entry_time_str,
				&fp->feed->entry->tm)) {
		stop(fp);
		return;
	}
	fp->feed->entry->t = mktime(&fp->feed->entry->tm);
	if (fp->feed->entry->t == (time_t)-1) {
		stop(fp);
		return;
	}
	free(fp->feed_entry_time_str);
	fp->feed_entry_time_str = NULL;
	entry_restore(fp);
}

static void rss_item_restore(struct feed_parser *fp);

static void pubdate_restore(struct feed_parser *fp)
{
	if (rfc822_time_date_parse(fp->feed_entry_time_str,
				&fp->feed->entry->tm)) {
		stop(fp);
		return;
	}
	fp->feed->entry->t = mktime(&fp->feed->entry->tm);
	if (fp->feed->entry->t == (time_t)-1) {
		stop(fp);
		return;
	}
	free(fp->feed_entry_time_str);
	fp->feed_entry_time_str = NULL;
	rss_item_restore(fp);
}

static void rdf_item_restore(struct feed_parser *fp);

static void dcdate_restore(struct feed_parser *fp)
{
	if (w3cdtf_time_date_parse(fp->feed_entry_time_str,
				&fp->feed->entry->tm)) {
		stop(fp);
		return;
	}
	fp->feed->entry->t = mktime(&fp->feed->entry->tm);
	if (fp->feed->entry->t == (time_t)-1) {
		stop(fp);
		return;
	}
	free(fp->feed_entry_time_str);
	fp->feed_entry_time_str = NULL;
	rdf_item_restore(fp);
}

static void atom_link(struct feed_parser *fp, const char **attrs,
		struct feed_link **l, void (*restore)(struct feed_parser *))
{
	struct feed_link *new = feed_link_new_attrs(attrs);
	if (!new->href) {
		perrf("atom link lacks href attribute");
		stop(fp);
		return;
	}
	new->next = *l;
	*l = new;
	skip(fp, restore);
}

static void rss_link(struct feed_parser *fp, struct feed_link **l,
		void (*restore)(struct feed_parser *))
{
	struct feed_link *new = feed_link_new();
	new->next = *l;
	*l = new;
	leaf(fp, &new->href, restore);
}

#define rdf_link rss_link

static void entry_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	switch (atom_tag_lex(name)) {
		case ATOM_TAG_TITLE:
			if (fp->feed->entry->title) {
				perrf("more than one title in Atom entry");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->title, entry_restore);
			break;
		case ATOM_TAG_SUMMARY:
			if (fp->feed->entry->summary) {
				perrf("more than one summary in Atom entry");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->summary, entry_restore);
			break;
		case ATOM_TAG_ID:
			if (fp->feed->entry->id) {
				perrf("more than one ID in Atom entry");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->id, entry_restore);
			break;
		case ATOM_TAG_LINK:
			atom_link(fp, attrs, &fp->feed->entry->link,
					entry_restore);
			break;
		case ATOM_TAG_UPDATED:
			/* Use updated tag
			 * only if published tag not used. */
			if (fp->feed->entry->t == (time_t)-1)
				leaf(fp, &fp->feed_entry_time_str,
						entry_time_restore);
			else
				skip(fp, entry_restore);
			break;
		case ATOM_TAG_PUBLISHED:
			leaf(fp, &fp->feed_entry_time_str,
					entry_time_restore);
			break;
		default:
			skip(fp, entry_restore);
			break;
	}
}

static void atom_restore(struct feed_parser *fp);

static void entry_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->entry->title) {
		perrf("title missing in Atom entry");
		stop(fp);
		return;
	}
	if (!fp->feed->entry->id) {
		perrf("ID missing in Atom entry");
		stop(fp);
		return;
	}
	if (fp->feed->entry->t == (time_t)-1) {
		perrf("updated time missing in Atom entry");
		stop(fp);
		return;
	}
	atom_restore(fp);
}

static void entry_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, entry_start, entry_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void feed_entry_new_prepend(struct feed *f)
{
	struct feed_entry *entry = feed_entry_new();
	entry->next = f->entry;
	entry->feed = f;
	f->entry = entry;
	f->len++;
}

static void entry(struct feed_parser *fp)
{
	feed_entry_new_prepend(fp->feed);
	entry_restore(fp);
}

static void rss_item_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	switch (rss_tag_lex(name)) {
		case RSS_TAG_TITLE:
			if (fp->feed->entry->title) {
				perrf("more than one title in RSS item");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->title, rss_item_restore);
			break;
		case RSS_TAG_DESCRIPTION:
			if (fp->feed->entry->summary) {
				perrf("more than one description in RSS item");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->summary, rss_item_restore);
			break;
		case RSS_TAG_GUID:
			if (fp->feed->entry->id) {
				perrf("more than one GUID in RSS item");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->id, rss_item_restore);
			break;
		case RSS_TAG_LINK:
			rss_link(fp, &fp->feed->entry->link,
					rss_item_restore);
			break;
		case RSS_TAG_PUBDATE:
			leaf(fp, &fp->feed_entry_time_str,
					pubdate_restore);
			break;
		default:
			skip(fp, rss_item_restore);
			break;
	}
}

static void rss_channel_restore(struct feed_parser *fp);

static void rss_item_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->entry->title && !fp->feed->entry->summary) {
		perrf("both title and description missing in RSS item");
		stop(fp);
		return;
	}
	if (!fp->feed->entry->id) {
		fp->feed->entry->id = strdup(
				fp->feed->entry->title ?
				fp->feed->entry->title :
				fp->feed->entry->summary);
		if (!fp->feed->entry->id)
			abort();
	}
	rss_channel_restore(fp);
}

static void rss_item_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rss_item_start, rss_item_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rss_item(struct feed_parser *fp)
{
	feed_entry_new_prepend(fp->feed);
	rss_item_restore(fp);
}

static void rdf_item_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	switch (rdf_tag_lex(name)) {
		case RDF_TAG_TITLE:
			if (fp->feed->entry->title) {
				perrf("more than one title "
						"in RDF site summary item");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->title, rdf_item_restore);
			break;
		case RDF_TAG_DESCRIPTION:
			if (fp->feed->entry->summary) {
				perrf("more than one description "
						"in RDF site summary item");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->entry->summary, rdf_item_restore);
			break;
		case RDF_TAG_LINK:
			if (fp->feed->entry->link) {
				perrf("more than one link "
						"in RDF site summary item");
				stop(fp);
				return;
			}
			rdf_link(fp, &fp->feed->entry->link,
					rdf_item_restore);
			break;
		case RDF_TAG_DATE:
			leaf(fp, &fp->feed_entry_time_str,
					dcdate_restore);
			break;
		default:
			skip(fp, rdf_item_restore);
			break;
	}
}

static void rdf_restore(struct feed_parser *fp);

static void rdf_item_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->entry->title) {
		perrf("title missing in RDF site summary item");
		stop(fp);
		return;
	}
	if (!fp->feed->entry->link) {
		perrf("link missing in RDF site summary item");
		stop(fp);
		return;
	}
	rdf_restore(fp);
}

static void rdf_item_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rdf_item_start, rdf_item_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rdf_item(struct feed_parser *fp, const char **attrs)
{
	const char *about = NULL;
	feed_entry_new_prepend(fp->feed);
	while (*attrs) {
		const char *a = ns_test(attrs[0], rdf_ns);
		if (a && strcmp(a, "about") == 0)
			about = attrs[1];
		attrs += 2;
	}
	if (!about) {
		perrf("about attribute missing in RDF site summary item");
		stop(fp);
		return;
	}
	fp->feed->entry->id = uri_escape(about);
	rdf_item_restore(fp);
}

static void atom_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	switch (atom_tag_lex(name)) {
		case ATOM_TAG_ENTRY:
			entry(fp);
			break;
		case ATOM_TAG_TITLE:
			if (fp->feed->title) {
				perrf("more than one title in Atom feed");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->title, atom_restore);
			break;
		case ATOM_TAG_LINK:
			atom_link(fp, attrs, &fp->feed->link, atom_restore);
			break;
		case ATOM_TAG_ID:
			if (fp->feed->id) {
				perrf("more than one ID in Atom feed");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->id, atom_restore);
			break;
		default:
			skip(data, atom_restore);
			break;
	}
}

static void atom_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->title) {
		perrf("title missing in Atom feed");
		stop(fp);
		return;
	}
	if (!fp->feed->id) {
		perrf("id missing in Atom feed");
		stop(fp);
		return;
	}
	if (!fp->feed->uri) {
		struct feed_link *l = fp->feed->link;
		while (l) {
			if (l->rel == FEED_LINK_REL_SELF) {
				fp->feed->uri = strdup(fp->feed->link->href);
				if (!fp->feed->uri)
					abort();
				break;
			}
			l = l->next;
		}
	}
}

static void atom_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, atom_start, atom_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void atom(struct feed_parser *fp)
{
	atom_restore(fp);
}

static void rss_channel_start(void *data, const char *name,
		const char **attrs)
{
	struct feed_parser *fp = data;
	switch (rss_tag_lex(name)) {
		case RSS_TAG_ITEM:
			rss_item(fp);
			break;
		case RSS_TAG_TITLE:
			if (fp->feed->title) {
				perrf("more than one title in RSS feed");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->title, rss_channel_restore);
			break;
		case RSS_TAG_LINK:
			rss_link(fp, &fp->feed->link, rss_channel_restore);
			break;
		case RSS_TAG_GUID:
			if (fp->feed->id) {
				perrf("more than one GUID in RSS feed");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->id, rss_channel_restore);
			break;
		default:
			skip(data, rss_channel_restore);
			break;
	}
}

static void rss_restore(struct feed_parser *fp);

static void rss_channel_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->title) {
		perrf("title missing in RSS feed");
		stop(fp);
		return;
	}
	if (!fp->feed->link) {
		perrf("link missing in RSS feed");
		stop(fp);
		return;
	}
	if (!fp->feed->uri)
		fp->feed->uri = strdup(fp->feed->link->href);
	if (!fp->feed->id)
		fp->feed->id = strdup(fp->feed->uri);
	rss_restore(fp);
}

static void rss_channel_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rss_channel_start, rss_channel_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rss_channel(struct feed_parser *fp)
{
	rss_channel_restore(fp);
}

static void rdf_channel_restore(struct feed_parser *fp);

static void rdf_channel_start(void *data, const char *name,
		const char **attrs)
{
	struct feed_parser *fp = data;
	switch (rdf_tag_lex(name)) {
		case RDF_TAG_TITLE:
			if (fp->feed->title) {
				perrf("more than one title "
						"in RDF site summary");
				stop(fp);
				return;
			}
			leaf(fp, &fp->feed->title, rdf_channel_restore);
			break;
		case RDF_TAG_LINK:
			if (fp->feed->link) {
				perrf("more than one link "
						"in RDF site summary "
						"channel");
				stop(fp);
				return;
			}
			rdf_link(fp, &fp->feed->link, rdf_channel_restore);
			break;
		default:
			skip(data, rdf_channel_restore);
			break;
	}
}

static void rdf_restore(struct feed_parser *fp);

static void rdf_channel_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	if (!fp->feed->title) {
		perrf("title missing in RDF site summary");
		stop(fp);
		return;
	}
	if (!fp->feed->link) {
		perrf("link missing in RDF site summary");
		stop(fp);
		return;
	}
	if (!fp->feed->uri) {
		fp->feed->uri = strdup(fp->feed->link->href);
		if (!fp->feed->uri)
			abort();
	}
	rdf_restore(fp);
}

static void rdf_channel_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rdf_channel_start, rdf_channel_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rdf_channel(struct feed_parser *fp, const char **attrs)
{
	const char *about = NULL;
	while (*attrs) {
		const char *a = ns_test(attrs[0], rdf_ns);
		if (a && strcmp(a, "about") == 0)
			about = attrs[1];
		attrs += 2;
	}
	if (!about) {
		perrf("about attribute missing in RDF site summary");
		stop(fp);
		return;
	}
	fp->feed->id = uri_escape(about);
	rdf_channel_restore(fp);
}

static void rss_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	if (rss_tag_lex(name) == RSS_TAG_CHANNEL) {
		/* According to the specification,
		 * there must be a single channel.
		 * Note that rss_channel_end checks
		 * if a title was indeed present;
		 * therefore, we can use this to determine
		 * if we already encountered a channel. */
		if (fp->feed->title) {
			perrf("more than one channel in rss tag");
			stop(fp);
			return;
		}
		rss_channel(fp);
	} else {
		perrf("unexpected child inside rss tag");
		stop(fp);
		return;
	}
}

static void rss_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	/* Note that rss_channel_end always checks for a title. */
	if (!fp->feed->title) {
		perrf("no channel tag found inside rss tag");
		stop(fp);
	}
}

static void rss_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rss_start, rss_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rss(struct feed_parser *fp)
{
	rss_restore(fp);
}

static void rdf_start(void *data, const char *name, const char **attrs)
{
	struct feed_parser *fp = data;
	enum RDF_TAG t = rdf_tag_lex(name);
	if (t == RDF_TAG_CHANNEL) {
		/* According to the specification,
		 * there must be a single channel.
		 * Note that rdf_channel always makes sure
		 * that an about attribute was indeed present;
		 * therefore, we can use this to determine
		 * if we already encountered a channel. */
		if (fp->feed->id) {
			perrf("more than one channel inside RDF tag");
			stop(fp);
			return;
		}
		rdf_channel(fp, attrs);
	} else if (t == RDF_TAG_ITEM) {
		rdf_item(fp, attrs);
	} else {
		skip(fp, rdf_restore);
	}
}

static void rdf_end(void *data, const char *name)
{
	struct feed_parser *fp = data;
	/* Note that rdf_channel always adds an id or stops parsing. */
	if (!fp->feed->id) {
		perrf("channel missing in RDF site summary");
		stop(fp);
		return;
	}
	if (!fp->feed->entry) {
		perrf("no items in RDF site sumarry");
		stop(fp);
		return;
	}
}

static void rdf_restore(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, rdf_start, rdf_end);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

static void rdf(struct feed_parser *fp)
{
	rdf_restore(fp);
}

static void root_start(void *data, const char *name, const char **attrs)
{
	if (atom_tag_lex(name) == ATOM_TAG_FEED)
		atom(data);
	else if (rss_tag_lex(name) == RSS_TAG_RSS)
		rss(data);
	else if (rdf_tag_lex(name) == RDF_TAG_RDF)
		rdf(data);
	else {
		perrf("XML root element not recognized");
		stop(data);
		return;
	}
}

static void root(struct feed_parser *fp)
{
	XML_SetElementHandler(fp->p, root_start, NULL);
	XML_SetCharacterDataHandler(fp->p, NULL);
}

/* fp->p is assumed to be already initialized. */
static void feed_parser_init2(struct feed_parser *fp)
{
	fp->err = 0;
	fp->feed = feed_new();
	fp->leaf_buf = NULL;
	fp->leaf_len = 0;
	fp->feed_entry_time_str = NULL;
	XML_SetUserData(fp->p, fp);
	root(fp);
}

/* fp->p is assumed to not be initialized yet. */
static void feed_parser_init(struct feed_parser *fp)
{
	fp->p = XML_ParserCreateNS(NULL, '\xff');
	if (!fp->p)
		abort();
	feed_parser_init2(fp);
}

/* Deinitialize all but fp->p. */
static void feed_parser_deinit2(struct feed_parser *fp)
{
	feed_free(fp->feed);
	free(fp->leaf_buf);
	free(fp->feed_entry_time_str);
}

static void feed_parser_deinit(struct feed_parser *fp)
{
	feed_parser_deinit2(fp);
	XML_ParserFree(fp->p);
}

/* Reset parser.
 * After the reset, the caller may reuse the parser as if newly allocated.
 * This function may be called at any moment,
 * but usually will be called after feed_parser_finish. */
void feed_parser_reset(struct feed_parser *fp)
{
	if (XML_ParserReset(fp->p, NULL) != XML_TRUE)
		abort();
	feed_parser_deinit2(fp);
	feed_parser_init2(fp);
}

struct feed_parser *feed_parser_new()
{
	struct feed_parser *fp = malloc(sizeof *fp);
	if (!fp)
		abort();
	feed_parser_init(fp);

	return fp;
}

void feed_parser_free(struct feed_parser *fp)
{
	feed_parser_deinit(fp);
	free(fp);
}

/* Copy and parse buf of size n.
 * Caller must indicate the last buffer with final non-zero. */
int feed_parser_parse(struct feed_parser *fp, char *buf, size_t n, int final)
{
	if (!XML_Parse(fp->p, buf, n, final)) {
		perrf("error parsing");
		return 1;
	}

	return 0;
}

/* Obtain buffer of size n to be filled by the caller
 * and later parsed with feed_parser_buf_parse.
 * Using this function avoids the copy which feed_parser_parse does. */
char *feed_parser_buf(struct feed_parser *fp, size_t n)
{
	char *buf;
	if (!(buf = XML_GetBuffer(fp->p, n)))
		abort();

	return buf;
}

/* Parse contents of buffer passed via feed_parse_buf. */
int feed_parser_buf_parse(struct feed_parser *fp, size_t n, int final)
{
	if (!XML_ParseBuffer(fp->p, n, final)) {
		perrf("error parsing");
		fp->err = 1;
		return 1;
	}

	return 0;
}

/* After caller already called feed_parser_buf_parse or feed_parser_parse
 * with argument final non-zero,
 * this function will return the finished feed.
 * Call this function at any other point in time
 * shall cause undefined behaviour.
 * If an error had been previously encountered,
 * this function returns NULL instead. */
struct feed *feed_parser_finish(struct feed_parser *fp)
{
	struct feed *res;
	if (fp->err)
		res = NULL;
	else {
		res = fp->feed;
		fp->feed = NULL;
	}

	return res;
}

struct feed *feed_file_parse(int fd)
{
	ssize_t n;
	char *buf;
	struct feed_parser fp;
	struct feed *res = NULL;

	feed_parser_init(&fp);
	do {
		buf = feed_parser_buf(&fp, BUF_SIZ);
		if ((n = read(fd, buf, BUF_SIZ)) < 0) {
			perrf_errno("Error reading feed");
			goto err;
		}
		if (feed_parser_buf_parse(&fp, n, n == 0))
			break;
	} while (n);

	res = feed_parser_finish(&fp);
	feed_parser_deinit(&fp);
	return res;

err:
	feed_parser_deinit(&fp);
	return NULL;
}

static int expat_check()
{
	const XML_Feature *li;
	size_t i;
	int end;

	li = XML_GetFeatureList();
	end = 0;
	for (i = 0; !end; i++) {
		switch (li[i].feature) {
			case XML_FEATURE_END:
				end = 1;
				break;
			case XML_FEATURE_SIZEOF_XML_CHAR:
				if (li[i].value != 1) {
					perrf(
						"Error: "
						"expat compiled with "
						"sizeof(XML_Char) "
						"different from 1"
					);
					abort();
				}
				break;
			default:
				break;
		}
	}

	return 0;
}

int feed_init()
{
	if (expat_check())
		return 1;
	if (regcomp(&xsd_time_date_preg, xsd_time_date_reg, REG_EXTENDED))
		return 1;
	if (regcomp(&rfc822_time_date_preg, rfc822_time_date_reg,
				REG_ICASE | REG_EXTENDED))
		return 1;
	if (regcomp(&w3cdtf_time_date_preg, w3cdtf_time_date_reg,
				REG_EXTENDED))
		return 1;
	if (utc_diff_init())
		return 1;

	return 0;
}
