// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <stdlib.h>
#include <stdint.h>
#include <limits.h>
#include <string.h>
#include "hashtab.h"

#define FNV_HASH_64_PRIME	0x00000100000001B3
#define FNV_HASH_64_OFFSET	0xCBF29CE484222325

uint64_t hash64(const char *s)
{
	uint64_t h = FNV_HASH_64_OFFSET;
	while (*s) {
		h ^= *s;
		h *= FNV_HASH_64_PRIME;
		s++;
	}

	return h;
}

#define FNV_HASH_32_PRIME	0x01000193
#define FNV_HASH_32_OFFSET	0x811C9DC5

uint32_t hash32(const char *s) {
	uint32_t h = FNV_HASH_32_OFFSET;
	while (*s) {
		h ^= *s;
		h *= FNV_HASH_32_PRIME;
		s++;
	}

	return h;
}

static uintmax_t hashtab_hash1_31(const char *s, unsigned short n)
{
	const uint32_t h = hash32(s);

	return ((h >> n) ^ h) & ((1 << n) - 1);
}

static uintmax_t hashtab_hash32(const char *s, unsigned short n)
{
	return hash32(s);
}

static uintmax_t hashtab_hash33_to_63(const char *s, unsigned short n)
{
	const uint64_t h = hash64(s);

	return ((h >> n) ^ h) & ((1 << n) - 1);
}

static uintmax_t hashtab_hash64(const char *s, unsigned short n)
{
	return hash64(s);
}

static unsigned short bits(uintmax_t n)
{
	unsigned short i;
	for (i = 0; i < sizeof n * CHAR_BIT && n >> i; i++)
		;

	return i;
}

static uintmax_t (*hash_func(unsigned short bits))
(const char *, unsigned short)
{
	if (bits < 32)
		return hashtab_hash1_31;
	else if (bits == 32)
		return hashtab_hash32;
	else if (bits < 64)
		return hashtab_hash33_to_63;
	else if (bits == 64)
		return hashtab_hash64;
	else
		return NULL;
}

struct hashtab *hashtab_new(size_t siz)
{
	size_t i;
	struct hashtab *res = malloc(sizeof *res);
	if (!res)
		abort();
	if (siz == 0)
		res->bits = 12;
	else
		res->bits = bits(siz - 1);
	res->siz = 1 << res->bits;
	res->hash = hash_func(res->bits);
	if (!res->hash)
		abort();
	res->tab = malloc(res->siz * sizeof *res->tab);
	if (!res->tab)
		abort();
	for (i = 0; i < res->siz; i++)
		res->tab[i] = NULL;
	res->len = 0;
	res->list = NULL;

	return res;
}

static uintmax_t hashtab_hash(struct hashtab *ht, const char *key)
{
	return ht->hash(key, ht->bits);
}

void *hashtab_ref(struct hashtab *ht, const char *key)
{
	struct hashtab_list *list = ht->tab[hashtab_hash(ht, key)];
	while (list) {
		if (strcmp(list->kv->key, key) == 0)
			return list->kv->val;
		list = list->next;
	}

	return NULL;
}

static void hashtab_grow(struct hashtab *ht)
{
	uintmax_t i;
	struct hashtab_list *a, *b;
	struct hashtab_key_val_list *c;
	/* Note that we only call this function when ht->len == ht->siz, so
	 * looping over ht->tab is not slower than looping over ht->list. */
	for (i = 0; i < ht->siz; i++) {
		a = ht->tab[i];
		while (a) {
			b = a->next;
			free(a);
			a = b;
		}
	}
	free(ht->tab);
	ht->siz = 1 << ++(ht->bits);
	ht->tab = malloc(ht->siz * sizeof *ht->tab);
	if (!ht->tab)
		abort();
	for (i = 0; i < ht->siz; i++)
		ht->tab[i] = NULL;
	ht->hash = hash_func(ht->bits);
	if (!ht->hash)
		abort();
	for (c = ht->list; c; c = c->next) {
		i = hashtab_hash(ht, c->key);
		a = malloc(sizeof *a);
		if (!a)
			abort();
		a->kv = c;
		a->next = ht->tab[i];
		ht->tab[i] = a;
	}
}

/* Assumes key not yet present.
 * Caller must pass a precomputed hash value for key. */
static inline void hashtab_add(struct hashtab *ht, const char *key,
		void *val, uintmax_t h)
{
	struct hashtab_key_val_list *kv;
	struct hashtab_list *hl;

	kv = malloc(sizeof *kv);
	if (!kv)
		abort();
	kv->key = strdup(key);
	if (!kv->key)
		abort();
	kv->val = val;
	kv->next = ht->list;
	kv->prev = NULL;
	ht->list = kv;

	hl = malloc(sizeof *hl);
	if (!hl)
		abort();
	hl->kv = kv;
	hl->next = ht->tab[h];
	hl->prev = NULL;
	ht->tab[h] = hl;

	ht->len++;
}

void *hashtab_set(struct hashtab *ht, const char *key, void *val)
{
	void *res;
	uintmax_t h = hashtab_hash(ht, key);
	struct hashtab_list *hl = ht->tab[h];
	while (hl) {
		if (strcmp(hl->kv->key, key) == 0) {
			res = hl->kv->val;
			hl->kv->val = val;
			return res;
		}
		hl = hl->next;
	}

	if (ht->len == ht->siz) {
		hashtab_grow(ht);
		/* Hash function may have changed with the grow! */
		h = hashtab_hash(ht, key);
	}
	hashtab_add(ht, key, val, h);

	return NULL;
}

void *hashtab_update(struct hashtab *ht, const char *key,
		void *(*f)(void *, void *), void *a)
{
	void *res;
	uintmax_t h = hashtab_hash(ht, key);
	struct hashtab_list *hl = ht->tab[h];
	while (hl) {
		if (strcmp(hl->kv->key, key) == 0) {
			res = hl->kv->val;
			hl->kv->val = f(res, a);
			return res;
		}
		hl = hl->next;
	}

	if (ht->len == ht->siz) {
		hashtab_grow(ht);
		/* Hash function may have changed with the grow! */
		h = hashtab_hash(ht, key);
	}
	hashtab_add(ht, key, f(NULL, a), h);

	return NULL;
}

void *hashtab_remove(struct hashtab *ht, const char *key)
{
	void *res;
	const uintmax_t h = hashtab_hash(ht, key);
	struct hashtab_list *hl = ht->tab[h];
	while (hl) {
		if (strcmp(hl->kv->key, key) == 0) {
			res = hl->kv->val;
			if (hl->kv->prev)
				hl->kv->prev->next = hl->kv->next;
			if (hl->kv->next)
				hl->kv->next->prev = hl->kv->prev;
			free(hl->kv->key);
			free(hl->kv);
			if (hl->prev)
				hl->prev->next = hl->next;
			if (hl->next)
				hl->next->prev = hl->prev;
			free(hl);
			ht->len--;

			return res;
		}
		hl = hl->next;
	}

	return NULL;
}

void hashtab_free(struct hashtab *ht, void val_free(void *))
{
	uintmax_t h;
	struct hashtab_list *a, *b;
	struct hashtab_key_val_list *c = ht->list, *d;
	while (c) {
		d = c->next;
		if (c->key) {
			h = hashtab_hash(ht, c->key);
			a = ht->tab[h];
			while (a) {
				b = a->next;
				free(a->kv->key);
				a->kv->key = NULL;
				if (val_free)
					val_free(a->kv->val);
				free(a);
				a = b;
			}
		}
		free(c);
		c = d;
	}
}
