// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <stdlib.h>
#include <perrf.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <curl/curl.h>
#include "feed.h"
#include "cmd.h"
#include "conf.h"
#include "status.h"
#include "misc.h"

#define PROMPT "rfeed? "
#define DEFAULT_CMD "f *"

static const char usage[] = "rfeed"
	" [-f feeds]"
	" [-s status]"
	" [-v viewer]"
	" [-d filter]"
	" [-c cmd]..."
	" [uri]...";

struct feeds {
	struct feed *feed;
	struct feeds *next;
};

struct arg_cmd {
	char *cmd;
	struct arg_cmd *next;
};

struct feeds *feeds;
struct feed_entry **entries_index;
size_t entries_index_len;
struct feed_parser *fp;
struct arg_cmd *arg_cmds;
struct arg_cmd *arg_cmd_last;
CURL *curl;

static int feed_parser_init()
{
	if (!(fp = feed_parser_new()))
		return 1;

	return 0;
}

static void feed_parser_deinit()
{
	feed_parser_free(fp);
	fp = NULL;
}

static size_t curl_cb(char *buf, size_t siz, size_t n, void *data)
{
	if (feed_parser_parse(fp, buf, n, 0))
		return CURL_WRITEFUNC_ERROR;

	return n;
}

static int curl_init()
{
	if (curl_global_init(CURL_GLOBAL_ALL))
		return 1;
	if (!(curl = curl_easy_init()))
		return 1;
	if (curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_cb))
		return 1;

	return 0;
}

static void curl_deinit()
{
	curl_easy_cleanup(curl);
	curl = NULL;
	curl_global_cleanup();
}

static struct feed *feed_get(const char *uri)
{
	if (curl_easy_setopt(curl, CURLOPT_URL, uri))
		abort();
	if (curl_easy_perform(curl))
		return NULL;
	if (feed_parser_parse(fp, NULL, 0, 1))
		return NULL;
	struct feed * const res = feed_parser_finish(fp);
	feed_parser_reset(fp);

	return res;
}

static int feeds_prepend(const char *uri)
{
	struct feeds *new;
	if (!(new = malloc(sizeof *new)))
		abort();
	new->feed = feed_get(uri);
	if (!new->feed) {
		perrf("Failed to get feed from %s", uri);
		return 1;
	}
	/* Some types of feed do not mandate a URI. */
	if (!new->feed->uri)
		feed_uri_set(new->feed, uri);
	new->next = feeds;
	feeds = new;
	entries_index_len += new->feed->len;
	return 0;
}

static int feed_entry_cmp(const void *p1, const void *p2)
{
	const struct feed_entry *e1 = *(const struct feed_entry **)p1;
	const struct feed_entry *e2 = *(const struct feed_entry **)p2;
	return e1->t > e2->t;
}

/* Expects entries_index_len to already hold the total number of entries. */
static void entries_index_init()
{
	size_t i;
	struct feed_entry *e;
	struct feeds *f;

	entries_index = malloc((sizeof *entries_index) * entries_index_len);
	if (!entries_index)
		abort();

	i = 0;
	f = feeds;
	while (f) {
		e = f->feed->entry;
		while (e) {
			if (i >= entries_index_len)
				abort();
			status_not_ancient(e);
			entries_index[i++] = e;
			e = e->next;
		}
		f = f->next;
	}
	qsort(entries_index, entries_index_len,
			sizeof(struct feed_entry *), feed_entry_cmp);
}

static int uri_file_proc(const char *file)
{
	FILE *f;
	char *line = NULL;
	size_t n = 0;
	ssize_t m;
	char *p, *q;
	f = fopen(file, "r");
	if (!f) {
		perrf_errno("error opening file %s", file);
		return 1;
	}
	while ((m = getline(&line, &n, f)) >= 0) {
		p = line;
		while (*p && isspace(*p))
			p++;
		if (*p == '\0' || *p == '#')
			continue;
		q = line + m - 1;
		while (q > p && isspace(*q))
			q--;
		if (q == p)
			continue;
		q[1] = '\0';
		if (feeds_prepend(p))
			return 1;
	}
	if (ferror(f))
		abort();
	free(line);
	if (fclose(f))
		abort();

	return 0;
}

static void arg_cmds_append(const char *str)
{
	struct arg_cmd *c = malloc(sizeof *c);
	if (!c)
		abort();
	c->cmd = strdup(str);
	if (!c->cmd)
		abort();
	c->next = NULL;
	if (!arg_cmds)
		arg_cmds = c;
	else
		arg_cmd_last->next = c;
	arg_cmd_last = c;
}

static void arg_cmds_proc()
{
	struct arg_cmd *c = arg_cmds;
	while (c) {
		cmd(c->cmd, entries_index, entries_index_len);
		c = c->next;
	}
}

static int args_proc(int argc, char *argv[])
{
	int opt;
	size_t i;
	while ((opt = getopt(argc, argv, "f:s:v:c:d:")) != -1) {
		switch (opt) {
			case 'c':
				arg_cmds_append(optarg);
				break;
			case 'f':
				if (uri_file_proc(optarg))
					return 1;
				break;
			case 's':
				conf_status_file_set(optarg);
				break;
			case 'v':
				conf_viewer_set(optarg);
				break;
			case 'd':
				conf_summary_filter_set(optarg);
				break;
			default:
				perrf("usage: %s", usage);
				return 1;
		}
	}
	for (i = optind; i < argc; i++)
		if (feeds_prepend(argv[i]))
			return 1;

	return 0;
}

static void history_proc(const char *line)
{
	size_t i;
	for (i = 0; line[i]; i++) {
		if (!isspace(line[i])) {
			add_history(line);
			return;
		}
	}
}

char *completion(const char *str, int st)
{
	static size_t i;
	char *res;
	if (st == 0)
		i = 0;
	while (i < cmdn) {
		const char * const name = cmds[i++].name;
		if (strpref(str, name)) {
			res = strdup(name);
			if (!res)
				abort();
			return res;
		}
	}

	i = 0;
	return NULL;
}

static void interact()
{
	char *line;
	if (cmd(DEFAULT_CMD, entries_index, entries_index_len)
			!= CMD_RES_SUCCESS)
		abort();
	using_history();
	rl_completion_entry_function = completion;
	while ((line = readline(PROMPT))) {
		history_proc(line);
		const int res = cmd(line, entries_index, entries_index_len);
		free(line);
		if (res == CMD_RES_QUIT)
			return;
		if (res != CMD_RES_SUCCESS)
			perrf("command syntax error");
	}
	if (ferror(stdin) || ferror(stdout))
		abort();
}

static void input_proc()
{
	char *line = NULL;
	size_t n = 0;
	ssize_t m;
	while ((m = getline(&line, &n, stdin)) >= 0) {
		const int res = cmd(line, entries_index, entries_index_len);
		if (res == CMD_RES_QUIT)
			return;
		if (res != CMD_RES_SUCCESS)
			perrf("command syntax error");
	}
	if (ferror(stdin))
		abort();
	free(line);
}

static void env_proc()
{
	char *str;
	if ((str = getenv("RFEED_VIEWER")))
		conf_viewer_set(str);
	if ((str = getenv("RFEED_SUMMARY_FILTER")))
		conf_summary_filter_set(str);
}

int main(int argc, char *argv[])
{
	int res;
	perrf_head = argv[0];
	if (feed_init())
		return 1;
	if (feed_parser_init())
		return 1;
	if (curl_init())
		return 1;
	env_proc();
	res = args_proc(argc, argv);
	curl_deinit();
	if (res)
		return res;
	feed_parser_deinit();
	if (status_init())
		return 1;
	entries_index_init();
	if (arg_cmds)
		arg_cmds_proc();
	else if (isatty(0))
		interact();
	else
		input_proc();
	if (status_deinit())
		return 1;

	return 0;
}
