// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <string.h>
#include <perrf.h>

int strpref(const char *pref, const char *str)
{
	size_t i;
	for (i = 0; pref[i] && str[i]; i++)
		if (pref[i] != str[i])
			return 0;

	return pref[i] == '\0';
}

static int str_write(int fd, const char *str)
{
	ssize_t n;
	size_t m = 0;
	const size_t len = strlen(str);
	while (m < len) {
		n = write(fd, str + m, len - m);
		if (n < 0)
			return 1;
		m += n;
	}

	return 0;
}

int prog_exec(const char *prog, const char *arg, const char *input)
{
	int res = 0;
	char c;
	ssize_t n;
	int flags;
	int exec_pipe[2] = { -1, -1};
	int input_pipe[2] = { -1, -1};
	int status;
	pid_t pid = -1;

	if (pipe(exec_pipe)) {
		res = 1;
		goto end;
	}
	flags = fcntl(exec_pipe[1], F_GETFD);
	if (flags < 0) {
		res = 1;
		goto end;
	}
	flags |= FD_CLOEXEC;
	if (fcntl(exec_pipe[1], F_SETFD, flags) == -1) {
		res = 1;
		goto end;
	}
	if (pipe(input_pipe)) {
		res = 1;
		goto end;
	}
	switch ((pid = fork())) {
		case 0:
			if (close(exec_pipe[0]))
				abort();
			if (close(input_pipe[1]))
				abort();
			if (dup2(input_pipe[0], 0))
				abort();
			if (close(input_pipe[0]))
				abort();
			execlp(prog, prog, arg, (char *)NULL);
			c = 0;
			while ((n = write(exec_pipe[1], &c, 1)) != 1) {
				if (n == -1 && errno != EINTR)
					break;
			}
			abort();
		case -1:
			perrf_errno("error forking");
			abort();
		default:
			if (close(exec_pipe[1])) {
				exec_pipe[1] = -1;
				res = 1;
				goto end;
			}
			exec_pipe[1] = -1;
			if (close(input_pipe[0])) {
				input_pipe[0] = -1;
				res = 1;
				goto end;
			}
			input_pipe[0] = -1;
			while ((n = read(exec_pipe[0], &c, 1)) != 0) {
				if (n == -1) {
					if (errno != EINTR) {
						perrf_errno(
							"error reading "
							"from exec pipe");
						res = 1;
						goto end;
					}
				} else {
					perrf("failed to execute %s", prog);
					res = 1;
					goto end;
				}
			}
			if (close(exec_pipe[0])) {
				exec_pipe[0] = -1;
				res = 1;
				goto end;
			}
			exec_pipe[0] = -1;
			break;
	}
	if (input) {
		if (str_write(input_pipe[1], input)) {
			perrf_errno("error writing to standard input of %s",
					prog);
			res = 1;
			goto end;
		}
	}

end:
	for (int i = 0; i < 2; i++) {
		if (exec_pipe[i] != -1) {
			if (close(exec_pipe[i]))
				abort();
		}
		if (input_pipe[i] != -1) {
			if (close(input_pipe[i]))
				abort();
		}
	}
	if (pid != -1) {
		while (waitpid(pid, &status, 0) == -1) {
			if (errno != EINTR) {
				perrf_errno("error waiting for process");
				abort();
			}
		}
		if (!WIFEXITED(status)) {
			perrf("program exited abnormally");
			res = 1;
		} else if (WEXITSTATUS(status)) {
			perrf("program exited with non-zero status %d",
					WEXITSTATUS(status));
			res = 1;
		}
	}

	return res;
}
