// Copyright (C) 2022-2024 B.R.S. Roso.
// 
// Rfeed is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License version 3 as published by the Free
// Software Foundation.
// 
// This program is distributed in the hope that it be useful but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// Rfeed.  If not, see <https://www.gnu.org/licenses/>.

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <hashtab.h>
#include <sprintf2.h>
#include <perrf.h>
#include "status.h"
#include "conf.h"

#define STATUS_FILE_UNSEEN 'U'
#define STATUS_FILE_VIEWED 'V'

#define STATUS_HT_OLD		0x1
#define STATUS_HT_VIEWED	0x2
#define STATUS_HT_NOT_ANCIENT	0x4

#define HASHTAB_SIZE (1 << 13)

struct hashtab *ht;

/* Caller must free returned string! */
static inline char *status_key(const char *uri, const char *id)
{
	char *key = sprintf2("%s %s", uri, id);
	if (!key)
		abort();
	return key;
}

/* Caller must free returned string! */
static inline char *status_key2(struct feed_entry *entry)
{
	if (!entry->feed->uri || !entry->id)
		abort();
	return status_key(entry->feed->uri, entry->id);
}

int status_init()
{
	int res;
	FILE *f;
	uintptr_t s;
	char *line = NULL;
	char *p = NULL;
	char *c, *uri, *id, *key;
	size_t n = 0;
	ssize_t m;
	ht = hashtab_new(HASHTAB_SIZE);
	if (!conf_status_file)
		return 0;
	f = fopen(conf_status_file, "r");
	if (!f) {
		if (errno == ENOENT)
			return 0;
		perrf_errno("error opening file %s", conf_status_file);
		return 1;
	}
	res = 0;
	while ((m = getline(&line, &n, f)) >= 0) {
		c = strtok_r(line, " \t", &p);
		if (!c) {
			res = 1;
			goto end;
		}
		if (*c == '\n')
			continue;
		if (strlen(c) != 1) {
			res = 1;
			goto end;
		}
		if (*c == STATUS_FILE_UNSEEN)
			s = STATUS_HT_OLD;
		else if (*c == STATUS_FILE_VIEWED)
			s = STATUS_HT_OLD | STATUS_HT_VIEWED;
		else {
			res = 1;
			goto end;
		}
		uri = strtok_r(NULL, " \t", &p);
		if (!uri) {
			res = 1;
			goto end;
		}
		id = strtok_r(NULL, "\n", &p);
		if (!id) {
			res = 1;
			goto end;
		}
		key = status_key(uri, id);
		hashtab_set(ht, key, (void *)s);
		free(key);
	}
	if (ferror(f))
		abort();
end:
	if (res)
		perrf("malformed status file at %s", conf_status_file);
	if (fclose(f))
		abort();
	free(line);

	return res;
}

/* Not a full deinit.
 * The hashtable is not deallocated. */
int status_deinit()
{
	FILE *f;
	struct hashtab_key_val_list *l;
	uintptr_t s;
	if (!conf_status_file)
		return 0;
	f = fopen(conf_status_file, "w");
	if (!f) {
		perrf_errno("Error opening file %s", conf_status_file);
		return 1;
	}
	l = ht->list;
	while (l) {
		s = (uintptr_t)l->val;
		if (s & STATUS_HT_NOT_ANCIENT) {
			if (fprintf(f, "%c %s\n",
						s & STATUS_HT_VIEWED ?
						STATUS_FILE_VIEWED :
						STATUS_FILE_UNSEEN,
						l->key) < 0)
				abort();
		}
		l = l->next;
	}
	if (fclose(f))
		abort();

	return 0;
}

static void *status_flag_update_cb(void *s, void *f)
{
	return (void *)((uintptr_t)s | (uintptr_t)f);
}

static inline void status_flag(struct feed_entry *entry, uintptr_t f)
{
	char *key = status_key2(entry);
	hashtab_update(ht, key, status_flag_update_cb, (void *)f);
	free(key);
}

/* Caller must ensure that entry->feed->uri never contain spaces. */
void status_viewed(struct feed_entry *entry)
{
	status_flag(entry, STATUS_HT_VIEWED);
}

/* Caller must ensure that entry->feed->uri never contain spaces. */
void status_not_ancient(struct feed_entry *entry)
{
	status_flag(entry, STATUS_HT_NOT_ANCIENT);
}

/* Caller must ensure that entry->feed->uri never contain spaces. */
int status(struct feed_entry *entry)
{
	char *key = status_key2(entry);
	const uintptr_t s = (uintptr_t)hashtab_ref(ht, key);
	free(key);
	if (s & STATUS_HT_VIEWED)
		return STATUS_VIEWED;
	if (s & STATUS_HT_OLD)
		return STATUS_UNSEEN;
	return STATUS_NEW;
}
